"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const request = require("request");
const change_waiter_1 = require("./change-waiter");
const self_state_1 = require("./self-state");
const etc_response_1 = require("./etc-response");
const etc_value_1 = require("./etc-value");
const dispatcher_1 = require("./dispatcher");
const rx = require("rxjs");
class EtcdObserable {
    constructor(cfg) {
        this.connected = null;
        this.selfStateInActions = {};
        this.cfg = cfg;
    }
    static create(cfg) {
        return new EtcdObserable(cfg);
    }
    request(method, url, options = {}) {
        return rx.Observable.create((ob) => {
            this.connect().subscribe((c) => {
                if (!c.isOk()) {
                    this.cfg.log.error('Request-REJECT no valid connection');
                    ob.error(c);
                    return;
                }
                const mkUrl = `${c.url}${url}`;
                const req = this.rawRequest(method, mkUrl, options);
                const bodies = [];
                req.on('data', (data) => {
                    bodies.push(data.toString());
                });
                req.on('complete', (resp) => {
                    const body = bodies.join('');
                    try {
                        const jsbody = JSON.parse(body);
                        if (jsbody.errorCode) {
                            this.cfg.log.debug('request:complete:error:', jsbody);
                            ob.error({ respond: resp, err: jsbody });
                        }
                        else {
                            this.cfg.log.debug('request:complete:ok:', method, mkUrl, options, body, resp.headers, resp.statusCode);
                            ob.next(jsbody);
                            ob.complete();
                        }
                    }
                    catch (e) {
                        this.cfg.log.debug(`request:complete:catch:error:[${body}]:`, method, mkUrl, options, resp.headers, resp.statusCode, e);
                        ob.error({ respond: resp, body: body, e: e });
                    }
                });
                req.on('error', (e) => {
                    this.cfg.log.debug('request:error:', method, mkUrl, options, e);
                    ob.error(e);
                });
                // on(event: 'request', listener: (req: http.ClientRequest) => void): this;
                // on(event: 'response', listener: (resp: http.IncomingMessage) => void): this;
                // on(event: 'data', listener: (data: Buffer | string) => void): this;
                // on(event: 'complete', listener: (resp: http.IncomingMessage, body?: string | Buffer) => void): this;
            }, (e) => {
                this.cfg.log.debug('request:error:', method, options, e);
                ob.error(e);
            }, () => {
                // ob.complete();
            });
        });
        // try {
        //   return new Promise<any>((resolve, reject) => {
        //   });
        // } catch (e) {
        //   if (e.name == 'StatusCodeError') {
        //     return Promise.reject(e);
        //   }
        //   this.cfg.log.error('Reconnected:', typeof e, e.name, e);
        //   this.connected = null; // reconnect etcd
        //   return this.request(method, url, options);
        // }
    }
    rawRequest(method, url, options = {}) {
        let my = {
            method: method,
            timeout: this.cfg.reqTimeout,
            uri: url
        };
        options = Object.assign(my, options);
        return request(options);
    }
    urlParams(params, sep = '') {
        let paramsStr = '';
        for (let key in params) {
            paramsStr += `${sep}${key}=${params[key]}`;
            sep = '&';
        }
        return paramsStr;
    }
    bodyParams(params) {
        return {
            headers: {
                'Content-Type': 'application/x-www-form-urlencoded'
            },
            body: this.urlParams(params)
        };
    }
    buildKeyUri(base, key) {
        let url;
        let app = this.cfg.appId || '';
        if (app.length != 0) {
            app += '/';
        }
        if (this.cfg.clusterId) {
            url = `${base}/${app}${this.cfg.clusterId}/${key}`;
        }
        else {
            url = `${base}/${app}${key}`;
        }
        return url;
    }
    untilFirstConnect() {
        return rx.Observable.create((ob) => {
            const result = [];
            // console.log('untilFirstConnect:', this.cfg.urls);
            const merged = rx.Observable.merge(...this.cfg.urls.map(url => this.selfStat(url))).subscribe(a => {
                // console.log('untilFirstConnect:a:', a);
                result.push(a);
                if (a.isOk()) {
                    ob.next(result);
                    ob.complete();
                    merged.unsubscribe();
                }
                else if (result.length == this.cfg.urls.length) {
                    ob.next(result);
                    ob.complete();
                }
            }, null, () => {
                // console.log('untilFirstConnect:completed:');
                ob.next(result);
                ob.complete();
            });
        });
        // filter(a => a.isOk()).first();
        // return rx.Observable.create((observer: rx.Observer<SelfState[]>) => {
        //   let resolved = false;
        //   let result: SelfState[] = [];
        //   this.cfg.urls.forEach((url) => {
        //     this.selfStat(url).then((value) => {
        //       if (resolved) { return; }
        //       result.push(value);
        //       if (value.isOk()) {
        //         resolved = true;
        //         res(result);
        //       } else {
        //         if (result.length == this.cfg.urls.length) {
        //           resolved = true;
        //           res(result);
        //         }
        //       }
        //     });
        //   });
        // });
    }
    retryConnect(obs, retry) {
        // console.log('retryConnect:', retry);
        if (retry > this.cfg.retries) {
            obs.next(self_state_1.default.error('all', []));
            obs.complete();
            return;
        }
        this.untilFirstConnect().subscribe((ret) => {
            // console.log('retryConnect:untilFirstConnect:', retry, ret);
            let foundOk = ret.find((s2) => s2.isOk());
            if (foundOk) {
                this.connected = foundOk;
                obs.next(foundOk);
                obs.complete();
                return;
            }
            this.cfg.log.info('Retry-Connect:', this.connected, retry, this.cfg.retries, this.cfg.waitTime);
            setTimeout(() => this.retryConnect(obs, retry + 1), this.cfg.waitTime);
        });
    }
    connect() {
        if (!this.connected) {
            return rx.Observable.create((obs) => {
                this.retryConnect(obs, 0);
            });
        }
        else {
            return rx.Observable.create((obs) => {
                obs.next(this.connected);
                obs.complete();
            });
        }
    }
    clusterState() {
        // console.log('this.cfg.urls:', this.cfg.urls)
        return rx.Observable.zip(...this.cfg.urls.map((url) => this.selfStat(url)));
    }
    resolvSelfStateInActions(url, s2) {
        let local = this.selfStateInActions[url];
        delete this.selfStateInActions[url];
        local.slice(1).forEach((ssia) => {
            ssia.resolve(s2);
        });
        return s2;
    }
    selfStat(url) {
        // console.log('selfStat:', url);
        if (!this.selfStateInActions[url]) {
            this.selfStateInActions[url] = [];
        }
        let dispatcher = dispatcher_1.default.create();
        this.selfStateInActions[url].push(dispatcher);
        if (this.selfStateInActions[url].length > 1) {
            // console.log('selfStat-Double:', url);
            return dispatcher.subject;
        }
        return rx.Observable.create((obs) => {
            // console.log('selfStat-create:', url);
            const ret = this.rawRequest('GET', `${url}/v2/stats/self`);
            ret.on('error', (err) => {
                // console.log('selfStat-create:error:', url);
                obs.next(this.resolvSelfStateInActions(url, self_state_1.default.error(url, err)));
                obs.complete();
            });
            const bodies = [];
            ret.on('data', (data) => {
                bodies.push(data.toString());
            });
            ret.on('complete', (resp) => {
                const body = bodies.join('');
                // console.log('selfStat-create:complete:', url, body);
                // this.cfg.log.info('complete', body);
                obs.next(this.resolvSelfStateInActions(url, self_state_1.default.ok(url, body)));
                obs.complete();
            });
        });
    }
    keyAction(method, key, options = {}) {
        let uri = this.buildKeyUri('/v2/keys', key);
        this.cfg.log.debug('keyAction', method, uri, options);
        return rx.Observable.create((obs) => {
            this.request(method, uri, options).subscribe((ret) => {
                // console.log('keyAction:ret:');
                obs.next(etc_response_1.default.fromJson(ret));
                obs.complete();
            }, (err) => {
                // console.log('keyAction:err:');
                obs.next(etc_response_1.default.error(err));
                obs.complete();
            });
        });
    }
    update(key, ttl) {
        return this.keyAction('PUT', key, this.bodyParams({
            'ttl': ttl,
            'refresh': true,
            'prevExist': true
        }));
    }
    addQueue(key, val, ttl) {
        return this.keyAction('POST', key, this.bodyParams({ 'value': val, 'ttl': ttl }));
    }
    mkdir(key) {
        return this.keyAction('PUT', key, this.bodyParams({ dir: true }));
    }
    rmdir(key, params = {}) {
        return this.keyAction('DELETE', `${key}${this.urlParams(Object.assign({ dir: true }, params), '?')}`);
    }
    delete(key) {
        return this.keyAction('DELETE', key);
    }
    list(key, params = {}) {
        // console.log('list:', `${key}${this.urlParams(params, '?')}`)
        return rx.Observable.create((obs) => {
            this.keyAction('GET', `${key}${this.urlParams(params, '?')}`).subscribe((list) => {
                if (list.isErr()) {
                    obs.next(etc_value_1.default.error(list));
                }
                if (!list.node.dir) {
                    obs.next(etc_value_1.default.error('not a directory'));
                }
                else {
                    obs.next(etc_value_1.default.value(list.node.nodes));
                }
                obs.complete();
            });
        });
    }
    getRaw(key, params = {}, options = {}) {
        // console.log('get:', `${key}${this.urlParams(params, '?')}`)
        return this.keyAction('GET', `${key}${this.urlParams(params, '?')}`, options);
    }
    getString(key, params = {}) {
        return rx.Observable.create((obs) => {
            this.getRaw(key, params).subscribe((er) => {
                if (er.isErr()) {
                    obs.next(etc_value_1.default.error(er));
                }
                else {
                    obs.next(etc_value_1.default.value(er.node.value));
                }
                obs.complete();
            }, (err) => {
                obs.next(etc_value_1.default.error(err));
            });
        });
    }
    getJson(key, params = {}) {
        return rx.Observable.create((obs) => {
            this.getRaw(key, params).subscribe(ret => {
                if (ret.isErr()) {
                    obs.next(etc_value_1.default.error(ret));
                }
                else {
                    obs.next(etc_value_1.default.value(JSON.parse(ret.node.value)));
                }
            }, (err) => {
                obs.next(etc_value_1.default.error(err));
            }, () => obs.complete());
        });
    }
    setRaw(key, val) {
        return rx.Observable.create((obs) => {
            this.keyAction('PUT', key, this.bodyParams({ value: val }))
                .subscribe(ret => {
                obs.next(ret);
            }, err => {
                obs.next(etc_response_1.default.error(err));
            }, () => obs.complete());
        });
    }
    setJson(key, val) {
        return this.setRaw(key, JSON.stringify(val));
    }
    createChangeWaiter(path, params = {}, options = {}) {
        return new change_waiter_1.default(this, path, params, options);
    }
}
exports.EtcdObserable = EtcdObserable;
exports.default = EtcdObserable;
//# sourceMappingURL=etcd-observable.js.map